# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    evolutionary optimizaton of plasmonic nanostructure geometry for 
    directional scattering

"""
## --- limit blas to n threads, do this BEFORE loading numpy
import os
nthreads = 2
os.environ["MKL_NUM_THREADS"] = "{}".format(int(nthreads))
os.environ["NUMEXPR_NUM_THREADS"] = "{}".format(int(nthreads))
os.environ["OMP_NUM_THREADS"] = "{}".format(int(nthreads))


import numpy as np
import matplotlib.pyplot as plt


from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core

from pyGDM2.EO import models as eo_models
from pyGDM2.EO import problems as eo_problems
from pyGDM2.EO.core import do_eo


from PyGMO import algorithm


#==============================================================================
# Setup pyGDM part
#==============================================================================
## ---------- Setup structure
mesh = 'cube'
step = 40
#material = materials.dummy(2.0)    # material: constant and real dielectric function
material = materials.gold()        # material: gold
n1, n2 = 1.0, 1.0         # constant environment

## --- Empty dummy-geometry, will be replaced on run-time by EO trial geometries
geometry = []       

struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))



## ---------- Setup incident field
field_generator = fields.planewave        # planwave excitation
kwargs = dict(theta = [0.0])              # several polarizations
wavelengths = [800]                       # one single wavelength
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)



## ---------- Simulation initialization
sim = core.simulation(struct, efield)




#==============================================================================
# setup evolutionary optimization
#==============================================================================
## --- output folder and file-names
results_folder = 'eo_out'
results_suffix = 'test_eo'


## --- structure model and optimizaiton problem
model = eo_models.BlockModel(sim, 
                             30, 1,1,1, [-10,10],
                             forbidden=[], symmetric=False)

problem_dimension = model.get_dim()
problem = eo_problems.ProblemDirectivity(problem_dimension, model, nthreads=1)


## --- algorithm configuration
algo = algorithm.jde(gen=1, memory=True)

islands = 1             # number of islands ( > 1: multi-processing)
population = 40         # individuals per island


## --- stop criteria
max_time = 3600         # seconds
max_iter = 200          # max. iterations
max_nonsuccess = 20     # max. consecutive iterations without improvement





#==============================================================================
# run evolution
#==============================================================================
#do_eo(results_folder, results_suffix, problem)    # run with default params
do_eo(results_folder, results_suffix,
      problem, algo, 
      islands, population,
      max_time, max_iter, max_nonsuccess, 
      ## --- flags
      runtime_plotting=0,
      save_each_generation=False,
      verbose=True)






















# encoding: utf-8
"""
Example script, comparing pyGDM with Mie theory for a silicon sphere

"""
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core
from pyGDM2 import tools
from pyGDM2 import linear
from pyGDM2 import visu




## --- load pre-calculated Mie-data
wl_mie, qext_mie, qsca_mie = np.loadtxt("scat_mie_Si_D150nm.txt").T
qabs_mie = qext_mie - qsca_mie


#==============================================================================
# pyGDM setup
#==============================================================================

## --- Setup incident field
field_generator = fields.planewave
## log-interval spectrum (denser at low lambda):
wavelengths = np.exp(np.linspace(np.log(300), np.log(1000), 30))
kwargs = dict(theta = [0.0])
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)


## --- Setup geometry (sphere D=150nm in vacuum)
## note: scale_factor = 2.7 was used for the example in the doc. paper, leading
## to a simulation time in the order of 1-2 hours. For an example this is 
## rather long running time, so the discretization is strongly reduced in this 
## script. A finer discretization is needed to reproduce exact peak positions, 
## and particulary also to correctly obtain the some higher order modes.
## however, even a rather coarse discretization gives pretty good
## approximates for the spectra (apart some 20-30nm shift and 
## ~10% intensity deviations)
scale_factor = 1.4
step = 18.75/scale_factor
radius = 4.*scale_factor
geometry = structures.sphere(step, R=radius, mesh='hex', ORIENTATION=2)

material = materials.silicon()
n1, n2 = 1.0, 1.0     # vacuum env.

struct = structures.struct(step, geometry, material, n1,n2, 
                                   structures.get_normalization('hex'),
                                   with_radiation_correction=True)


sim = core.simulation(struct, efield)

visu.structure(sim)
print('(hex) ----- N_dipoles =', len(sim.struct.geometry), end='')

#%%
#==============================================================================
# run the simulation
#==============================================================================
E = core.scatter(sim, method='lu', verbose=True)
#E = core.scatter(sim, method='cupy', matrix_setup='numba', verbose=True)


## extinction spectrum
field_kwargs = tools.get_possible_field_params_spectra(sim)[0]
wl, spec = tools.calculate_spectrum(sim, field_kwargs, linear.extinct)
a_ext, a_sca, a_abs = spec.T
a_geo = tools.get_geometric_cross_section(sim)



#%%
#==============================================================================
# plot
#==============================================================================
plt.figure()
plt.title("silicon sphere, D=150nm")

## --- Mie
plt.plot(wl_mie, qext_mie, 'b--', dashes=[2,1],label='ext.')
plt.plot(wl_mie, qabs_mie, 'g--', dashes=[2,1],label='abs.')
plt.plot(wl_mie, qsca_mie, 'r--', dashes=[2,1],label='scat.')

## --- pyGDM
plt.plot(wl, a_ext/a_geo, linewidth=1., color='b', label='')
plt.scatter(wl, a_ext/a_geo, marker='x', linewidth=1.5, color='b', label='')
plt.scatter(wl, a_abs/a_geo, marker='x', linewidth=1.5, color='g', label='')
plt.scatter(wl, a_sca/a_geo, marker='x', linewidth=1.5, color='r', label='')

## --- for legend only
plt.plot([0], [0], 'k--', dashes=[2,1], label='Mie')
plt.scatter([0], [0], marker='x', linewidth=1.5, color='k', label='pyGDM')
## -- legend
plt.legend(loc='best', fontsize=12)


plt.xlabel("wavelength (nm)")
plt.ylabel("ext. / scat. / abs. efficiency")
plt.xlim( [wl.min(), wl.max()] )
plt.ylim( [0, 10] )



plt.tight_layout()
plt.show()



# encoding: utf-8
"""
tools for EO submodule of pyGDM2

    Copyright (C) 2017-2024, P. R. Wiecha

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    

"""
from __future__ import print_function
from __future__ import absolute_import

import six; from six.moves import cPickle as pickle

import os
import copy
import warnings

import numpy as np



def reload_eo(filename):
    """Reload saved evol. optimization from file
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
        
    Returns
    -------
    dict : evolutionary optimization data
    
    """
    if type(filename) in [str, np.str_]:
        try:
            eo_dict = pickle.load( open(filename, "rb") )
        except IOError:
            raise IOError("Saved optimization file not found!")
    elif type(filename) == dict:
        eo_dict = filename
    else:
        raise ValueError("No valid optimization data. Must be either string (path to file) or dict.")
        
    return eo_dict
    
def load_eo(filename):
    """wrapper for `reload_eo`"""
    return reload_eo(filename)
    


#==============================================================================
# generate simuation dictionary for best solution
#==============================================================================
def get_best_candidate(filename, iteration=-1, do_test=True, verbose=False):
    """generate simuation object for best solution
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
        
    iteration : int, default: -1
        iteration (Nr of improvements) to use. Best candidate of selected iteration
        will be returned. Python indexing, hence positive and negative numbers 
        possible.
    
    do_test : bool, default: True
        whether to test the stored fitness (redo the simulation) or not
    
    verbose : bool, default: False
        print some information
        
    Returns
    -------
    sim : :class:`pyGDM2.core.simulation` instance of best candidate
    """
    ## --- reload
    eo_dict = reload_eo(filename)
    
    ## --- champion of selected iteration
    halloffame = eo_dict['halloffame']
    champ = halloffame[iteration]
    f = champ[1]
    x = champ[2]
    
    ## --- generate geometry
    problem = eo_dict['problem']
    problem.model.generate_structure(x)
    
    if verbose:
        i_iter = champ[0]
        if iteration < 0:
            iteration = len(halloffame) + iteration
        print ("Best candidate after {} iterations (with {} improvements): fitness = {}".format(
                                    i_iter+1, iteration, ["{:.5g}".format(_f) for _f in f]))
    
    ## --- test
    if do_test:
        if verbose: print ("Testing: recalculating fitness...", end='')
        f_test = problem.fitness(x)
        differ_flag = False
        for f_test_1, f_test_2 in zip(f_test, f):
            f1 = round(f_test_1, 5)
            f2 = round(f_test_2, 5)
            if f1-f2 != 0:
                differ_flag = True
        
        if differ_flag:
            warnings.warn("Stored and recalculated fitness values seem to differ. " + 
                          "f_stored={}, f_recalc={}".format(
                                      ["{:.5g}".format(_f) for _f in f],
                                      ["{:.5g}".format(_f) for _f in f_test]))
        elif verbose: print ("Done. Everything OK.")
    
    ## --- copy and return simulation dict
    problem.model.generate_structure(x)
    sim = copy.deepcopy(problem.model.sim)
    return sim


def get_best_candidate_f_x(filename, iteration=-1):
    """get best fitness and parameter vector of a specific generation.
    
    Returns also the total number of generations with improvements accessible 
    by the `iteration` parameter.
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
        
    iteration : int, default: -1
        iteration (Nr of improvements) to use. Best candidate of selected iteration
        will be returned. Python indexing, hence positive and negative numbers 
        possible.
    
    Returns
    -------
    f : array like
        fitness(-vector) of best candidate
    
    x : array like
        parameter vector of best candidate
    
    N : int
        number of iterations with improvement
    
    """
    ## --- reload
    eo_dict = reload_eo(filename)
    
    ## --- champion of selected iteration
    halloffame = eo_dict['halloffame']
    champ = halloffame[iteration]
    f = champ[1]
    x = champ[2]
    
    return f, x, len(halloffame)
    

def get_model(filename):
    """get instance of model object used in optimization
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
    
    Returns
    -------
    model : Instance of class based on :class:`.model.BaseClass` 
        Model used in the loaded EO
    """
    eo_dict = reload_eo(filename)
    problem = eo_dict['problem']
    
    return problem.model
    

def get_problem(filename):
    """get instance of problem object used in optimization
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
    
    Returns
    -------
    model : Instance of class based on :class:`.problems.BaseProblem`
        Problem used in the loaded EO
    """
    eo_dict = reload_eo(filename)
    problem = eo_dict['problem']
    
    return problem
    

def get_population(filename):
    """get full final population
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
    
    Returns
    -------
    list of lists : list of tuples (f, x)
        Full population of final iteration. Each individual is stored as
        tuple (*f*, **x**) with *f* fitness the of parameter-vector **x**
    """
    eo_dict = reload_eo(filename)
    pop = eo_dict['pop']
    pop_list = [(pop.get_f()[i], pop.get_x()[i]) for i in range(len(pop))]
    
    return pop_list




def get_pareto_fronts(filename, iteration=-1, recalc=False, verbose=False):
    """get Pareto fronts and corresponding simulations of multi-objective optimization
    
    Parameters
    ----------
    filename : str or dict
        either path to filename (pickled optimization data) or the optimization
        dictionary. In the latter case, the input will simply be returned 
        unaltered.
    
    iteration : int, default: -1
        iteration number. Python indexing (pos. and neg. indices possible).
    
    Returns
    -------
    list of lists : list of tuples (f, x)
        Full population of final iteration. Each individual is stored as
        tuple (*f*, **x**) with *f* fitness the of parameter-vector **x**
    """
    import pygmo as pg
    
    eo_dict = reload_eo(filename)
    
    if iteration != -1 and eo_dict['all_gen_pop'] == []:
        raise ValueError("Population-evolution not stored. To obtain the " +
                         "Pareto-fronts for others than the final population, " +
                         "`save_all_generations` flag must be enabled " +
                         "during optimization. ")
    if iteration == -1:
        pop = eo_dict['pop']
    else:
        pop = eo_dict['all_gen_pop'][iteration][1]
    
    pf, dl, dc, ndl = pg.fast_non_dominated_sorting(pop.get_f())
    problem = eo_dict['problem']

    allpareto = []
    allsim = []
    allx = []
    for p in pf:
        allpareto.append([])
        allsim.append([])
        allx.append([])
        for i in p:
            best_x = pop.get_x()[i]
            best_f = pop.get_f()[i]
            problem.model.generate_structure(best_x)
            if recalc:
                problem._objfun_impl(best_x)
            
            allpareto[-1].append(best_f)
            allsim[-1].append( copy.deepcopy(problem.model.sim) )
            allx[-1].append( copy.deepcopy(best_x) )
            
        allpareto[-1] = np.transpose(allpareto[-1])
        allpareto[-1] *= -1
        allpareto[-1] = np.append(allpareto[-1],[np.arange(len(allpareto[-1][0]))], axis=0)
        allpareto[-1] = allpareto[-1].T
        
        ## sort pareto front and get sorting indices
        if np.shape(allpareto[-1])[0] != 1:    
            allpareto[-1] = np.sort(allpareto[-1].view('f8,f8,i8'), order=['f1'], axis=0).view(np.float32).T
            sort_idx = [int(i) for i in allpareto[-1][2]]
        else:
            ## one single element in Pareto front
            sort_idx = [0]
        allpareto[-1] = -1 * allpareto[-1][:-1].T
        
        ## Sort simulations and parameter vectors according to pareto-front order
        allsim[-1] = [allsim[-1][s_idx] for s_idx in sort_idx]
        allx[-1] = [allx[-1][s_idx] for s_idx in sort_idx]

    return allpareto, allsim, allx
    


def plot_pareto_2d(pareto, print_indices=True, cutoff_indexprint=-1, 
                   marker='x', mew=1.5, ms=7, 
                   show=True, **kwargs):
    """plot the pareto-front, optionally print index numbers for identification
    
    Parameters
    ----------
    pareto : list of tuples
        Pareto-front as list of objective-value tuples (2 objectives)
    
    print_indices : bool, default: True
        print index of individuals next to pareto-front points
    
    cutoff_indexprint : float, default: -1
        optional: manual value for minimum distance between two individuals in
        Pareto plot for allowing the printing of a new index-value label
    
    marker, mew, ms: str, float, float. defaults: 'x', 1.5, 7
        marker type, marker linewith and marker size for step-plot
    
    show : bool, default: True
        whether or not to run `plt.show()`
    
    **kwargs: are passed to matplotlib's `plt.step`
    
    Returns
    -------
    return value of matplotlib's `plt.step`
    """
    import matplotlib.pyplot as plt
    
    ## plot
    if pareto.shape[1] == 0:
        warnings.warn("Empty Pareto-front. Ignoring.")
    else:
        step_plot = plt.step(pareto.T[0], pareto.T[1], where='post',
                                     marker=marker, mew=mew, ms=ms, **kwargs)
        
        ## labels with indices
        if print_indices:
            if cutoff_indexprint == -1:
                cutoffx = (pareto.T[0].max() - pareto.T[0].min())*0.05  # 5% of objective #1 value range
                cutoffy = (pareto.T[1].max() - pareto.T[1].min())*0.05  # 5% of objective #1 value range
            else:
                cutoffx = cutoff_indexprint
                cutoffy = cutoff_indexprint
            
            x_last, y_last = 0, 0
            for i, (x,y) in enumerate(zip(pareto.T[0], pareto.T[1])):
                ## check whether to print or not
                if not (abs(x - x_last)<cutoffx and abs(y - y_last)<cutoffy):
                    plt.text(x + (pareto.T[0].max()-pareto.T[0].min())*0.0, 
                             y + (pareto.T[1].max()-pareto.T[1].min())*0.04, 
                             i)
                    x_last, y_last = x, y
        
        plt.xlabel('fitness objective #1')
        plt.ylabel('fitness objective #2')
        
        if show:
            plt.show()
        return step_plot


def plot_all_pareto_fronts_2d(all_paretos, cmap=None, show=True, **kwargs):
    """Plot all Pareto-fronts
    
    Parameters
    ----------
    all_paretos : list of lists of tuples
        List of all non-dominated fronts as obtained by 
        :func:`.get_pareto_fronts`
    
    cmap : matplotlib colormap, default: None
        A colormap for the different pareto-fronts. Default "None" will use 
        matplotlib's `Blues`
    
    show : bool, default: True
        whether or not to run `plt.show()`
    
    **kwargs: are passed to :func:`.plot_pareto_2d`
    
    """
    import matplotlib.pyplot as plt
    if cmap is None:
        cmap = plt.cm.Blues
    
    colors = cmap(np.linspace(0.3, 1.0, len(all_paretos)))[::-1]
    for i, pareto in enumerate(all_paretos):
        plot_pareto_2d(pareto, show=False, color=colors[i], 
                                   label='front #{}'.format(i), **kwargs)
    plt.legend(loc='best')

    if show:
        plt.show()




#==============================================================================
# Get index-lists for directionality problem
#==============================================================================
verbose=True
plot=True
plot_type='mesh' #'scatter'

def calculate_solid_angle_by_dir_index(Nteta, Nphi, tetamin, tetamax, dir_index,
                                      plot=True, show=True,
                                      plot_type='scatter', verbose=True):
    """calculate target solid angle for `pyGDM2.EO.problems.ProblemDirectivity`
    
    plot_type : str, default: "scatter"
        either "mesh", "scatter" or "both". 
        Note: Since the field is evaluated at points, 'mesh' is only an 
        approximation and represents the true solid angle not entirely correctly!
        
    """
    def theta_phi_lists(Nteta, Nphi, tetamin, tetamax):
        dteta = (tetamax-tetamin)/(Nteta-1.)
        dphi = 2.*np.pi/(Nphi) 
        
        tetalist = np.zeros( (Nteta, Nphi) )
        philist  = np.zeros( (Nteta, Nphi) )
        for j in range(Nteta):
            for k in range(Nphi):
                tetalist[j][k] = tetamin + j*dteta
                philist[j][k] = k*dphi
        
        return tetalist.flatten(), philist.flatten(), dteta, dphi
    
    
    if type(dir_index) == list:
        dir_index = np.array(dir_index)
    
    theta, phi, dTheta,dPhi = theta_phi_lists(Nteta, Nphi, tetamin, tetamax)
    dS = np.sin(theta)*dTheta*dPhi
    dS += dS.max()*0.1
    
    
    
    #==============================================================================
    # list of selected/unselected points
    #==============================================================================
    teta_phi_sel = []
    teta_phi_unsel = []
    for idx, tp in enumerate(zip(theta, phi)):
        if dir_index is not None:
            if idx in dir_index:
                teta_phi_sel.append(tp)
            else:
                teta_phi_unsel.append(tp)
    teta_phi_sel = np.transpose(teta_phi_sel)
    teta_phi_unsel = np.transpose(teta_phi_unsel)
    
    #==============================================================================
    # selection colormesh (0 --> not selected, 1--> selected)
    #==============================================================================
#    Nteta2 = Nteta+2
    Nteta2 = Nteta+1
    theta2, phi2,dTheta2,dPhi2 = theta_phi_lists(Nteta2, Nphi, tetamin, tetamax)
#    dir_index2 = dir_index + Nphi
    dir_index2 = dir_index
    
    ## --- create colormesh arrays
    t_p_colors = np.zeros( theta2.shape )
    t_p_colors[dir_index2] = 1
    
    shape = (Nteta2, Nphi)
    t_p_colors = t_p_colors.reshape(shape)
    t_p_teta = theta2.reshape(shape)
    t_p_phi  = phi2.reshape(shape) - (dPhi2/2.)
    
    ### --- for plotting only: Add 360degrees (copy of 0 degrees)
    t_p_teta   = np.concatenate([t_p_teta.T, [t_p_teta.T[-1]]]).T
    t_p_phi    = np.concatenate([t_p_phi.T, [np.ones(t_p_phi.T[-1].shape) * t_p_phi.max()+dPhi2]]).T #- np.pi/float(Nphi)
    t_p_colors = np.concatenate([t_p_colors.T, [t_p_colors.T[-1]]]).T
    
    
    #==============================================================================
    # Print index-list
    #==============================================================================
    if verbose:
        print ('    idx, theta, phi' )
        print ('-----------------------' )
        for i, tp in enumerate(zip(theta, phi)):
            if dir_index is not None:
                if i in dir_index:
                    print ("XXX -- ", end='')
                else:
                    print (" o  -- ", end='')
                
            print ('{: >3}:   t={:.1f}, p={:.1f},     dS={:.3f}'.format(i, tp[0] * (180./np.pi), tp[1] * (180./np.pi), dS[i]))
    
    
    #==============================================================================
    # PLOT
    #==============================================================================
    if plot:
        import matplotlib.pyplot as plt
        from matplotlib.colors import LinearSegmentedColormap
        cmap = LinearSegmentedColormap.from_list('mycmap', [(0,  (1.0, 0.7, 0.7) ),
                                                            (1,  (0.4, 0.7, 0.4) )])
        
        ## --- scatter
        plt.subplot(polar=True)
        
        if plot_type in ['mesh', 'both']:
            plt.pcolormesh(t_p_phi, t_p_teta*180/np.pi, t_p_colors, cmap=cmap, alpha=1, edgecolors="none")
        
        if plot_type in ['scatter', 'both']:
            plt.scatter(teta_phi_unsel[1], teta_phi_unsel[0]*180/np.pi, marker='x', color='r', label='unsel.')
            plt.scatter(teta_phi_sel[1], teta_phi_sel[0]*180/np.pi, marker='x', color='g', label='sel.', lw=3, s=35)
            
        plt.ylim(0, 90)
        if show:
            plt.show()
                
    
    
if __name__=='__main__':
    ##--- TESTING

    
    
    #==============================================================================
    # Configuration
    #==============================================================================
    dir_index = None
    tetamin = 0
    tetamax = np.pi/2.
    
    
    #==============================================================================
    # DONUT
    #==============================================================================
    Nteta = 7
    Nphi = 36
    dir_index = np.arange(144, 180) # donut --> ring at teta=58deg
    
    
    
    #==============================================================================
    # FOCUSED
    #==============================================================================
    Nteta = 7
    Nphi = 36
    dir_index = np.array([109,142,143]) # tight focus 1 - 44 window x3
    
    
    dir_index = np.array([143+36]) # tight focus 2 - 58s
    toffset  = 1*Nphi
    dir_index = np.array([109+toffset,142+toffset,143+toffset]) # tight focus 58 w3
    dir_index = np.array([143+72]) # tight focus 3 - 73s
    dir_index = np.array([143+108]) # tight focus 4 - 90s
    toffset  = 3*Nphi
    dir_index = np.array([109+toffset,142+toffset,143+toffset]) # tight focus 90 w3
    dir_index = np.array([109+toffset+17,143+toffset-18]) # tight focus 90 w3
    
    #toffset  = 1*Nphi; toffset2 = 2*Nphi
    #dir_index = np.array([109+toffset,142+toffset,143+toffset, 109+toffset2,142+toffset2,143+toffset2]) # tight focus 5 - 6pts window
    #
    #dir_index = np.array([109-36,110-36,111-36,112-36,139-36,140-36,141-36,142-36,143-36,
    #                     109,110,111,112,139,140,141,142,143,
    #                     109+36,110+36,111+36,112+36,139+36,140+36,141+36,142+36,143+36]) # loose focus
    
    
    #==============================================================================
    # DOUBLE FOCUS
    #==============================================================================
    #Nteta = 7
    #Nphi = 37
    #toffset  = 0*Nphi; toffset2 = 0
    #dir_index = np.array([149+toffset,183+toffset,184+toffset, 165+toffset2,166+toffset2,167+toffset2]) # double focus at 0deg / 180deg
    
    
    #==============================================================================
    # SIMPLE
    #==============================================================================
    Nteta = 3
    Nphi=5
    tetamin=0
    tetamax=np.pi/2.
    dir_index = [10]
    
    
    calculate_solid_angle_by_dir_index(Nteta, Nphi, tetamin, tetamax, dir_index, plot_type='scatter')    
    calculate_solid_angle_by_dir_index(Nteta, Nphi, tetamin, tetamax, dir_index, plot_type='mesh')    
    calculate_solid_angle_by_dir_index(Nteta, Nphi, tetamin, tetamax, dir_index, plot_type='both')    
    
    
    
    
    
    
    
    
    



# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha, A. Arbouet
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
fast electron simulations

Contributed by A. Arbouet, CEMES-CNRS, 2020
email: arbouet@cemes.fr

"""
from __future__ import print_function
from __future__ import absolute_import

import warnings
# import math

import numpy as np
import copy
import numba

import time

# =============================================================================
# additional visualization
# =============================================================================
def visu_structure_electron(struct, x0=None, y0=None, 
                            scale_e_pos=10, marker_e_pos='x', color_e_pos='r',
                            show=True, **kwargs):
    """plot 2D structure projection together with electron beam position
    
    all further kwargs are passed to `pyGDM2.visu.structure`
    
    Parameters
    ----------
    struct : list or :class:`.core.simulation`
          either list of 3d coordinate tuples or simulation description object 
          
    x0, y0 : float, default: None
        position of the electron beam in the (OXY) plane. If `None` and 
        simulation instance is passed as structure, try to infer electron 
        beam position(s) from incident field.
    
    scale_e_pos : float, default: 15
        scale of electron-position marker
    
    marker_e_pos : str, default: 'x'
        maker symbol for electron-position
    
    color_e_pos : str, default: 'r'
        color of electron-position marker
    
    Returns
    -------
    result returned by matplotlib's `scatter`
    
    """
    import matplotlib.pyplot as plt
    from pyGDM2 import visu
    from pyGDM2 import tools
    from pyGDM2 import core_py
    from pyGDM2 import fields_py
    step = tools.get_step_from_geometry(struct)
    
    ## if not given, try to infer electron positions defined in simulation
    if x0 is None and y0 is None and type(struct) == core_py.simulation:
        if struct.efield.field_generator.__name__ == fields_py.fast_electron.__name__:
            elpos = []
            for kw in struct.efield.kwargs_permutations:
                elpos.append([kw['x0'], kw['y0']])
            x0, y0 = np.unique(elpos, axis=0).T
    
    im = visu.structure(struct, show=False, **kwargs)
    plt.scatter(x0, y0, color=color_e_pos, marker=marker_e_pos, s=scale_e_pos*np.int32(step), lw=3)    
    plt.gca().set_aspect('equal')
    
    if show:
        plt.show()
    
    return im




# =============================================================================
# helper
# =============================================================================
def electron_speed(electron_kinetic_energy):
    """ Velocity of electron with kinetic energy electron_kinetic_energy
       
    Parameters
    ----------   
    electron_kinetic_energy : float
        electron kinetic energy (keV)
    
    Returns
    -------
    velec : speed of electron in units of the speed of light
    """        
    Eo_el = 511.0   # electron rest mass (keV)
    gamma = 1.0 + electron_kinetic_energy/Eo_el 
    beta = np.sqrt(1.0 - 1.0/gamma**2)
    return beta





# =============================================================================
# evaluation functions
# =============================================================================
def CL(sim, field_index, r=10000., tetamin=0, tetamax=np.pi, Nteta=30, 
       phimin=0, phimax=2*np.pi, Nphi=36, polarizerangle='none', 
       return_value='int_Es'):
    """Cathodoluminescence emission generated by a fast electron moving along (OZ)
       
    The fundamental E-field created by the fast electron is given by the 
    function :func:`pyGDM2.fields.fast_electron`.
    Electron kinetic energy, beam position and propagation direction are parameters 
    of fields.fast_electron.
    
    Using :func:`pyGDM2.linear_py.farfield` for calculation of farfield scattering.
    
    For further documentation on the possible kwargs, their meaning and 
    returned values, see doc of :func:`pyGDM2.linear_py.farfield`
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
        
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
           
    Returns
    -------    
    CL_signal : Cathodoluminescence emission intensity 
        number of emitted photons per eV per electron
        
    Notes
    -----
    For details of the EELS computation:
    Arnaud Arbouet et al. *Electron energy losses and cathodoluminescence 
    from complex plasmonic nanostructures: spectra, maps and radiation patterns 
    from a generalized field propagator*, New J. Phys. 16, 113012 (2014)
    """
    from pyGDM2 import linear_py as linear
        
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    field_params = sim.E[field_index][0]
    wavelength   = field_params['wavelength'] 
    MULT = 2.31E20
    factor = MULT * wavelength
    
    ff_scat = linear.farfield(sim, field_index, r_probe=None, r=r, 
                              tetamin=tetamin, tetamax=tetamax, Nteta=Nteta, 
                              phimin=phimin, phimax=phimax, Nphi=Nphi, 
                              polarizerangle=polarizerangle, return_value=return_value)
    
    if return_value.lower() in ["map", "efield"]:        
        CL_signal = [ff_scat[0], ff_scat[1], 
                     ff_scat[2]*factor, ff_scat[3]*factor, ff_scat[4]*factor]
    else:
        CL_signal = factor * ff_scat
 
    return CL_signal



def EELS(sim, field_index):       
    """Electron Energy Loss probability of a fast electron moving along (OZ)
       
    The E-field created by the fast electron is given by the function fields.fast_electron.
    Electron kinetic energy, beam position and propagation direction are parameters 
    of fields.fast_electron.       
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
        
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
           
    Returns
    -------    
    eels_signal : electron energy loss probability per eV per electron
        
    Notes
    -----
    For details of the EELS computation:
    Arnaud Arbouet et al 2014 New J. Phys. 16 113012     
    """
    from pyGDM2 import tools
    
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. " +
                         "Run `core.scatter` simulation first.")
    
    ## --- constants
    MULT =  929448800             # Multiplication constant 
    
    step = tools.get_step_from_geometry(sim)
    field_params = sim.E[field_index][0]
    wavelength   = field_params['wavelength']    
    electron_kinetic_energy  = field_params['electron_kinetic_energy'] 
    x0    = field_params['x0']   
    y0    = field_params['y0']
    
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    eps2 = env_dict['eps2']
    n2 = eps2**0.5
    
    k0 =  n2*(2.*np.pi/wavelength)  
    velec = electron_speed(electron_kinetic_energy)
    krho = k0*np.sqrt(1-velec**2.) / velec
    
# =============================================================================
#     Computation of E-field along electron trajectory
# =============================================================================
    ## compute the dipole moments at each discretization cell
    Eint = sim.E[field_index][1]
    alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    P = np.matmul(alpha_tensor, Eint[...,None])[...,0]
       
    ## compute the loss probability
    from scipy import special
    res = 1j*0.
    for idip, _pos in enumerate(sim.struct.geometry):   
        RT = np.sqrt( (_pos[0] - x0)**2. + (_pos[1] - y0)**2. )
        if (RT > 1.5*step):
            U = krho * RT
            contrib1 = - krho**2.*special.kv(0, U) * P[idip][2]
            contrib2 = 1j*k0*krho*((_pos[0]-x0) * P[idip][0] + 
                                   (_pos[1]-y0) * P[idip][1]) * \
                                              special.kv(1, U) / (velec*RT)
            contrib = (contrib1 + contrib2) * np.exp(-1j*k0*_pos[2]/velec)
        else: 
            contrib = 0
        res = res + contrib
    
    ## electron energy loss
    eels_ret = np.real(MULT*res/k0)
        
    return eels_ret







# encoding: utf-8
"""
Example script, demonstrating raster-scan simulations

"""
from __future__ import print_function, division

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core
from pyGDM2 import linear
from pyGDM2 import visu
from pyGDM2 import tools

import numpy as np
import matplotlib.pyplot as plt




# =============================================================================
# Simulation setup
# =============================================================================
## --- Setup structure
mesh = 'cube'
step = 7.0   # discretization in nm
L = 21       # cube side-length in nm
geometry = structures.rect_wire(step, L=L/step,H=L/step, W=L/step, mesh=mesh)
geometry.T[2] += step/2.

material = materials.dummy(2.0); MATERIAL = '20'
n1, n2 = 1.0, 1.0  # constant environment

struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))
struct = structures.center_struct(struct)



## --- Setup incident field (including dipole positions to evaluate)
x0 = np.linspace(-250, 250, 25)                  # x-scan pos.
y0 = np.linspace(-250, 250, 25)                  # y-scan pos.
z0 = struct.geometry.T[2].max() + step/2. + 15   # fixed height: 15nm above struct

kwargs = dict(x0=x0, y0=y0, z0=z0,    # positions where to evaluate
              mx=0,my=0,mz=0)         # dipole orientation (0,0,0) --> placeholder
wavelengths = [500]

## -- field-instances for electric and magnetic dipoles
efield_p = fields.efield(fields.dipole_electric, wavelengths, kwargs)
efield_m = fields.efield(fields.dipole_magnetic, wavelengths, kwargs)

## ---------- Simulation initialization
sim_p = core.simulation(struct, efield_p)
sim_m = core.simulation(struct, efield_m)


#%%
# =============================================================================
# Run simulation
# =============================================================================
print("\nEvaluation electric dipole decay rate..." )
SBB_p = core.decay_rate(sim_p, method='lu', verbose=True)

print("\nEvaluation magnetic dipole decay rate..." )
SBB_m = core.decay_rate(sim_m, method='lu', verbose=True)


#%%
# =============================================================================
# Plot decay for several orientations of the e-/m-dipole
# =============================================================================

## list of dipole test orientations
dp_list = [[1, 0, 0], 
           [0, 1, 0], 
           [0, 0, 1]]


plt.figure(figsize=(8,8))
for i, (mx,my,mz) in enumerate(dp_list):

    SBB_map_e = linear.decay_eval(sim_p, SBB_p[0], mx,my,mz, verbose=1)
    SBB_map_m = linear.decay_eval(sim_m, SBB_m[0], mx,my,mz, verbose=1)
    
    
    plt.subplot(3,2,2*i+1, aspect='equal')
    plt.title("{} dipole || ({},{},{})".format('electric',mx,my,mz))
    visu.scalarfield(SBB_map_e, cmap='jet', show=False)
    plt.colorbar(label='gamma / gamma_0')
    
    plt.subplot(3,2,2*i+2, aspect='equal')
    plt.title("{} dipole || ({},{},{})".format('magnetic',mx,my,mz))
    visu.scalarfield(SBB_map_m, cmap='jet', show=False)
    plt.colorbar(label='gamma / gamma_0')
    

plt.tight_layout()
plt.show()











from __future__ import print_function
from __future__ import absolute_import

import os
import sys
import platform
import re

from setuptools import setup



# =============================================================================
# helper functions to extract meta-info from package
# =============================================================================
def read_version_file(*parts):
    return open(os.path.join(*parts), 'r').read()

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def find_version(*file_paths):
    version_file = read_version_file(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

def find_name(*file_paths):
    version_file = read_version_file(*file_paths)
    version_match = re.search(r"^__name__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find name string.")

def find_author(*file_paths):
    version_file = read_version_file(*file_paths)
    version_match = re.search(r"^__author__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find author string.")



print ("\n" + 60*'#' + "\n")

# =============================================================================
# fortran modules: with or without openmp support
# =============================================================================
def check_openmp():
    """
    compile with OpenMP ?
    """
    if "WITH_OPENMP" in os.environ and os.environ["WITH_OPENMP"] == "False":
        print ("No OpenMP requested by environment")
        return False

    if ("--no-openmp" in sys.argv):
        sys.argv.remove("--no-openmp")
        os.environ["WITH_OPENMP"] = "False"
        print ("No OpenMP requested by command line")
        return False
    
    if ("--with-openmp" in sys.argv):
            sys.argv.remove("--with-openmp")
            os.environ["WITH_OPENMP"] = "True"
            print ("OpenMP explicitly requested by command line")
            return True
        
    if platform.system() == 'Darwin':
        print ("Running OSX: OpenMP not implemented in default 'clang'. \nIf you encounter problems, you might try '--no-openmp' to compile without OpenMP")
        return True
    
    print ("Compiling fortran module with OpenMP.\n"+
           "  - use '--with-fortran' to install with the deprecated fortran API (`pyGDM2.f_api`)." +
           "  - use '--no-openmp' to deactivate openmp.\n")
    return True



if "--with-fortran" in sys.argv:
    if check_openmp():
        openmp = "-fopenmp" 
        openmp_linker = "-lgomp"
    else:
        openmp = "" 
        openmp_linker = ""



# =============================================================================
# with or without fortran modules?
# =============================================================================
package_list=['pyGDM2', 'pyGDM2.fields', 'pyGDM2.propagators', 'pyGDM2.multipole', 'pyGDM2.EO']
    
    
print ("\n" + 60*'#' + 2*"\n")



# =============================================================================
# main setup    
# =============================================================================
setup(
    name = find_name("pyGDM2", "__init__.py"),
    version = find_version("pyGDM2", "__init__.py"),
    author = find_author("pyGDM2", "__init__.py"),
    author_email = "pwiecha@laas.fr",
    description = ("A python full-field electrodynamical solver, "
                   "based on the Green dyadic method (volume integral technique "
                   "in frequency domain)."),
    license = "GPLv3+",
    long_description=read('README.rst'),
    packages=package_list,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Physics",
        "Environment :: Console",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Intended Audience :: Science/Research"
    ],
    url = 'https://gitlab.com/wiechapeter/pyGDM2',
    download_url = '',
    keywords = ['coupled dipoles method', 'green dyadic method', 'electrodynamical simulations', 'nano optics', 'frequency-domain'],
    install_requires=['numpy', 'numba'],
    python_requires='>=3.7',
) 



import matplotlib.pyplot as plt

from pyGDM2 import fields
from pyGDM2 import tools
from pyGDM2 import visu


field_generator = fields.evanescent_planewave
kwargs = dict(theta_inc=45, polar='p', returnField='E')



## --- layered environment
n3 = 1.0            # vacuum
n2 = 1.05 + 1.8j    # lossy interface layer (metallic)
spacing = 40.0      # thickness of interface layer n2 (in nm)
n1 = 1.5            # dielectric substrate

## --- 2D evaluation volume (plane)
projection = 'XZ'
r_probe = tools.generate_NF_map(-500,500,30, -400,600,30,0, projection=projection)

## -- evaluation of the field-generator
wavelength = 500
NF = tools.evaluate_incident_field(field_generator, wavelength, kwargs, r_probe,
                            n1=n1,n2=n2,n3=n3, spacing=spacing)


## -- helper function to plot the interface layer
def plot_layer():
    plt.axhline(0, color='w',lw=1); plt.axhline(0, dashes=[2,2], color='k',lw=1)
    plt.axhspan(0,spacing, color='k', ls='--', fc='orange', alpha=0.35)
    plt.axhline(spacing, color='w',lw=1); plt.axhline(spacing, dashes=[2,2], color='k',lw=1)
    plt.arrow(350, -350, -150, 150, head_width=50, head_length=80, fc='k', ec='k')




plt.figure(figsize=(12,5))
plt.subplot(141, aspect='equal')
v = visu.vectorfield(NF, complex_part='real', projection=projection, tit=projection+' real part', show=0)
plot_layer()

plt.subplot(142, aspect='equal')
v = visu.vectorfield(NF, complex_part='imag', cmap=plt.cm.Reds, projection=projection, tit=projection+' imag part',  show=0)
plot_layer()


plt.subplot(243, aspect='equal')
v = visu.vectorfield_color(NF, projection=projection, tit='intensity: '+r'$|E|^2/|E0|^2$', show=0)
plt.colorbar(v)
plt.xticks([]); plt.yticks([])
plot_layer()

plt.subplot(244, aspect='equal')
v = visu.vectorfield_color(NF, projection=projection, fieldComp='ex', tit='Ex', show=0)
plt.colorbar(v)
plt.xticks([]); plt.yticks([])
plot_layer()

plt.subplot(247, aspect='equal')
v = visu.vectorfield_color(NF, projection=projection, fieldComp='ey', tit='Ey', show=0)
plt.colorbar(v)
plt.xticks([]); plt.yticks([])
plot_layer()

plt.subplot(248, aspect='equal')
v = visu.vectorfield_color(NF, projection=projection, fieldComp='ez', tit='Ez', show=0)
plt.colorbar(v)
plt.xticks([]); plt.yticks([])
plot_layer()

plt.show()



# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
"""
Collection of incident fields
"""

from __future__ import print_function
from __future__ import absolute_import

import itertools
import warnings
import copy
import types
import cmath

import numpy as np
import numba




#==============================================================================
# globals
#==============================================================================
DTYPE_C = np.complex64



#==============================================================================
# Incident field container class
#==============================================================================
class efield(object):
    """incident electromagnetic field container class
    
    Defines an incident electric field including information about wavelengths,
    polarizations, focal spot or whatever optional parameter is supported by
    the used field generator.
    
    Parameters
    ----------
    field_generator : `callable`
        field generator function. Mandatory arguments are 
          - `struct` (instance of :class:`.structures.struct`)
          - `wavelength` (list of wavelengths)
    
    wavelengths : list
        list of wavelengths (wavelengths in nm)
    
    kwargs : list of dict or dict
        possible additional keyword arguments, passed to `field_generator`.
        Either dict or list of dicts.
          - If list of dicts, each entry must correspond exactly to one 
            parameters-set for `field-generator`.
          - If dict, maybe contain lists for configurations of the parameters. 
            In that case, all possible parameter-permutations will be generated.
    
    Examples
    --------
    >>> kwargs = dict(theta = [0.0,45,90])
    [{'theta': 0.0}, {'theta': 45.0}, {'theta': 90.0}]
    
    is equivalent to:
    
    >>> kwargs = [dict(theta=0.0), dict(theta=45.0), dict(theta=90.0)]
    [{'theta': 0.0}, {'theta': 45.0}, {'theta': 90.0}]
    
    """
    def __init__(self, field_generator, wavelengths, kwargs):
        """initialize field container"""
        self.field_generator = field_generator
        self.wavelengths = np.array(wavelengths)
        self.kwargs = copy.deepcopy(kwargs)
        _kwargs = copy.deepcopy(kwargs)
        
        ## --- generate parameter-sets for field-generator
        if type(_kwargs) == dict:
            ## --- integer parameters to list
            for key in _kwargs:
                if type(_kwargs[key]) not in [list, np.ndarray]:
                    _kwargs[key] = [_kwargs[key]]
            
            ## --- generate all permutations of kwargs for direct use in field_generator
            varNames = sorted(_kwargs)
            self.kwargs_permutations = [dict(zip(varNames, prod)) for 
                                    prod in itertools.product(*(_kwargs[varName] 
                                                for varName in varNames))]
        elif type(_kwargs) == list:
            self.kwargs_permutations = []
            for kw in _kwargs:
                self.kwargs_permutations.append(kw)
                if type(kw) != dict:
                    raise ValueError("Wrong input for 'kwargs': Must be either dict or list of dicts.")
                
        else:
            raise ValueError("Wrong input for 'kwargs': Must be either dict or list of dicts.")
        
        ## set precision to single by default
        self.setDtype(np.float32, np.complex64)
    
    
    def setDtype(self, dtypef, dtypec):
        """set dtype of arrays"""
        self.dtypef = dtypef
        self.dtypec = dtypec
    
    
    ## !!! Todo: add automatic field gradient calculation
    # def get_field_gradient():
    #     from pyGDM2 import linear
    #     gradE0 = linear.field_gradient(sim, field_index, r_qe, 
    #                                    delta=eps_dd, which_fields=['E0'])
    #     gradH0 = linear.field_gradient(sim, field_index, r_qm, 
    #                                    delta=eps_dd, which_fields=['H0'])


    def __repr__(self, verbose=False):
        out_str =  ' ----- incident field -----'
        out_str += '\n' + '   field generator: "{}"'.format(self.field_generator.__name__)
        out_str += '\n' + '   {} wavelengths between {} and {}nm'.format(
                         len(self.wavelengths), self.wavelengths.min(),
                         self.wavelengths.max(),)
        if verbose or len(self.wavelengths)<6:
            for i, wl in enumerate(self.wavelengths):
                out_str += '\n' + '      - {}: {}nm'.format(i,wl)
        out_str += '\n' + '   {} incident field configurations per wavelength'.format(
                                        len(self.kwargs_permutations))    
        if verbose or len(self.kwargs_permutations)<6:
            for i, kw in enumerate(self.kwargs_permutations):
                out_str += '\n' + '      - {}: {}'.format(i,str(kw).replace("{","").replace("}",""))
        return out_str







# encoding: utf-8
"""
Example script, simulating the coupling of a dipolar emitter to a 
gold splitring resonator.

"""
from __future__ import print_function, division

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core
from pyGDM2 import linear
from pyGDM2 import tools
from pyGDM2 import visu

import numpy as np
import matplotlib.pyplot as plt






#==============================================================================
# setup simulation
#==============================================================================
mesh = 'cube'
step = 30.0
geometry = structures.rect_split_ring(step, L1=12,L2=16,H=3,W=3, G=False, mesh=mesh)
material = materials.gold()

n1, n2 = 1.0, 1.0  # constant environment

geometry = structures.center_struct(geometry)
struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))



## --- Setup incident field
field_generator = fields.dipole_electric   # light-source: dipolar emitter
kwargs = dict(x0=0, y0=0, z0=1.5*step, 
              mx=[1,0], my=[0,1], mz=0)
wavelengths = [1000.]
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)



## --- Simulation initialization
sim = core.simulation(struct, efield)

tools.print_sim_info(sim, verbose=1)






#%%
#==============================================================================
# run the simulation
#==============================================================================
E = core.scatter(sim, method='LU', verbose=True)
#E = core.scatter(sim, method='cuda', matrix_setup='numba', verbose=True)
#E = core.scatter(sim, method='cuda', matrix_setup='cuda', verbose=True)


## --- farfield pattern ("backfocal plane image")
Nteta=40; Nphi=2*72

teta, phi, I_sc_0, I_tot_0, I0_0 = linear.farfield(
                                sim, field_index=0,   # index 0: dipole || X 
                                tetamin=0, tetamax=np.pi/2., 
                                Nteta=Nteta, Nphi=Nphi)
teta, phi, I_sc_90, I_tot_90, I0_90 = linear.farfield(
                                sim, field_index=3,   # index 3: dipole || Y
                                tetamin=0, tetamax=np.pi/2., 
                                Nteta=Nteta, Nphi=Nphi)





#%%
#==============================================================================
# plot
#==============================================================================
scale_I = 1E16

def conf_polar():
    plt.ylim([0, 90])
    plt.gca().set_xticklabels([])
    plt.gca().set_yticklabels([])


plt.figure()

## --- structure geometry
plt.subplot(131, aspect='equal')
plt.title("structure")
plt.axis('off')

## structure
visu.structure(geometry, color='#c7a800', scale=1, show=False)

bbox = plt.gca().get_window_extent().transformed(plt.gcf().dpi_scale_trans.inverted())
width, height = bbox.width*plt.gcf().dpi, bbox.height*plt.gcf().dpi
               
## dipole position
plt.text(0, 40, r"$\mathbf{p}$", ha='center', va='bottom')
plt.scatter([0], [0], marker='x', linewidth=2, s=20, color='k')

## scale bar
plt.text(120,-330, "100nm", ha='center', va='bottom')
plt.plot([70,170] , [-350,-350], lw=2, color='k', clip_on=False)

geo_dim = 300
plt.xlim([-geo_dim, geo_dim])
plt.ylim([-geo_dim, geo_dim])


## --- pattern for dipole along X
plt.subplot(132, polar=True)
plt.title("$\mathbf{p} \parallel X$", x=0.5, y=1.15)
im = visu.farfield_pattern_2D(teta, phi, I_sc_0*scale_I, degrees=True, show=False)
conf_polar()

plt.colorbar(im, orientation='horizontal', label=r'$|\mathbf{E}|^2$ (a.u.)', shrink=0.75, 
             pad=0.05, aspect=15, ticks=[0, 0.5, 1.0])
plt.clim([0, 1])


## --- pattern for dipole along Y
plt.subplot(133, polar=True)
plt.title("$\mathbf{p} \parallel Y$", x=0.5, y=1.15)
im = visu.farfield_pattern_2D(teta, phi, I_sc_90*scale_I, degrees=True, show=False)
conf_polar()

plt.colorbar(im, orientation='horizontal', label=r'$|\mathbf{E}|^2$ (a.u.)', shrink=0.75, 
             pad=0.05, aspect=15, ticks=[0, 0.05, 0.10, 0.15])
plt.clim([0, 0.15])


plt.tight_layout()
plt.show()






"""
pyGDM2.EO - evolutionary optimization toolkit for use with pyGDM2. 


    Copyright (C) 2017-2024, P. R. Wiecha

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    


Additional requirements :
    - pyGMO2/paGMO2 (https://esa.github.io/pagmo2/)


"""
__name__ = 'pyGDM2.EO'
__author__ = 'Peter R. Wiecha'

__all__ = ["EO.core", "EO.models", "EO.problems", "EO.tools"]

## make modules available
import pyGDM2.EO.core
import pyGDM2.EO.models
import pyGDM2.EO.problems
import pyGDM2.EO.tools



# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    Demonstrate how to continue a formerly ran evolutionary optimizaton. 
    Continuing the example-optimization of plasmonic nanostructure geometry for 
    directional scattering.

"""


from pyGDM2.EO.core import continue_eo



## --- optimization file locations
results_folder = 'eo_out'
results_suffix = 'test_eo'


## --- Continue evolution
continue_eo(results_folder, results_suffix,
            runtime_plotting=False,
            save_each_generation=False)






















# encoding: utf-8
#
#Copyright (C) 2017-2023, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
python implementations of the Green's Dyads, accelerated using `numba`
"""
from __future__ import print_function
from __future__ import absolute_import

import copy
import warnings

import numpy as np
import math
import cmath

import numba
# from numba import cuda

# =============================================================================
# Base `dyads` class
# =============================================================================
### --- dyad class 3D non-retarded
class DyadsBaseClass(object):
    """
    Bases: object
    
    Green's tensors container class - pure-python API
    
    Defines the frame of a linear GDM simulation via the Green's tensors,
    self-terms and material-polarizability relations.
    
    Parameters
    ----------
        
    dtype : (optional) `str`, default: 'f'
        precision ('f': float/single/32bit or 'd': double/64bit)
    
    """
    __name__ = "Base Green's tensors class"
    
    def __init__(self, dtype='f'):
        if dtype=='f':
            self.dtypef = np.float32
            self.dtypec = np.complex64
        else:
            self.dtypef = np.float64
            self.dtypec = np.complex128
    
        ## Dyads
        self.G_EE = self.notImplemented
        self.G_HE = self.notImplemented
        self.G_EE_ff = self.notImplemented
        
        ## repropagator routine
        self.repropagate = self.notImplemented
        
        ## coupling matrix constructor routines
        self.tsbs_EE = self.notImplemented
        self.tsbs_HE = self.notImplemented
        
    
    def __repr__(self):
        """description about simulation environment defined by set of dyads
        """
        out_str =  ' ------ NULL Tensors -------'
        return out_str
    
    
    def _legacyStructCompatibility(self, struct):
        if (self.n1_material is None or self.n2_material is None or 
            self.n3_material is None or self.spacing is None):
            ## !!! use `struct` object will be deprecated in the future!
            warnings.warn("Deprecation warning: " +
                          "Environment should be specified in the `dyads` object. " +
                          "Falling back to definition from `struct` class. " +
                          "This will raise an Exception in future versions of pyGDM.")
            if hasattr(struct, 'n1_material'):
                self.spacing = struct.spacing
                self.n1_material = struct.n1_material
                self.n2_material = struct.n2_material
                self.n3_material = struct.n3_material
            else:
                raise Exception("Structure does not contain environment definition. " +
                                "Please specify the environment in the Green's tensor class.")
    
    
    def notImplemented(self, *args, **kwargs):
        raise NotImplementedError("This feature is not implemented in this Green's tensor set.")
    
    
    def exceptionHandling(self, struct, efield):
        """Exception handling / consistency check for the set of tensors"""
        warnings.warn("Dummy exception handling. Should be implemented in child class.")
        return True
    
    
    def getConfigDictG(self, wavelength, struct, efield):
        """return numba typed-dict which is used by Green's tensor functions"""
        raise NotImplementedError("Definition missing! This function needs to be overridden in child class!")
    
    
    def getEnvironmentIndices(self, wavelength, geo):
        """return list of encironment permittivity at each meshpoint location"""
        raise NotImplementedError("Definition missing! This function needs to be overridden in child class!")
        
        
    def getSelfTermEE(self, wavelength, struct):
        """return list of 'EE' self-term tensors (3x3) at each meshpoint"""
        raise NotImplementedError("Definition missing! This function needs to be overridden in child class!")

        
    def getSelfTermHE(self, wavelength, struct):
        """return list of 'HE' self-term tensors (3x3) at each meshpoint"""
        raise NotImplementedError("Definition missing! This function needs to be overridden in child class!")
        
        
    def getPolarizabilityTensor(self, wavelength, struct):
        """return list of polarizability tensors (3x3) at each meshpoint"""
        raise NotImplementedError("Definition missing! This function needs to be overridden in child class!")


### --- non retarded 3D
# =============================================================================
# numba compatible propagators (CPU + CUDA), 3D
# =============================================================================
## --- free space propagator
@numba.njit(cache=True)
def _G0(R1, R2, wavelength, eps):
    """
    R1: dipole position
    R2: evaluation position
    """
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] - R1[2]
    lR = math.sqrt(Dx**2 + Dy**2 + Dz**2)
    
    k = 2*np.pi / wavelength
    cn = cmath.sqrt(eps)
    ck0 = 1j * k * cn
    k2 = k*k*eps
    
    r25 = math.pow((Dx*Dx+Dy*Dy+Dz*Dz), 2.5)
    r2 = math.pow((Dx*Dx+Dy*Dy+Dz*Dz), 2.0)
    r15 = math.pow((Dx*Dx+Dy*Dy+Dz*Dz), 1.5)
    
#!C-------------------------------------------------------------------
    T1XX = -1*(Dy*Dy+Dz*Dz) / r15
    T2XX = (2*Dx*Dx-Dy*Dy-Dz*Dz) / r2
    T3XX = (2*Dx*Dx-Dy*Dy-Dz*Dz) / r25
#!C-------------------------------------------------------------------
    T1XY = Dx*Dy / r15
    T2XY = 3*Dx*Dy / r2
    T3XY = 3*Dx*Dy / r25
#!C-------------------------------------------------------------------
    T1XZ = Dx*Dz / r15
    T2XZ = 3*Dx*Dz / r2
    T3XZ = 3*Dx*Dz / r25
#!C-------------------------------------------------------------------
    T1YY = -(Dx*Dx+Dz*Dz) / r15
    T2YY = (2*Dy*Dy-Dx*Dx-Dz*Dz) / r2
    T3YY = (2*Dy*Dy-Dx*Dx-Dz*Dz) / r25
#!C-------------------------------------------------------------------
    T1YZ = Dy*Dz / r15
    T2YZ = 3*Dy*Dz / r2
    T3YZ = 3*Dy*Dz / r25
#!C------------------------------------------------------------------
    T1ZZ = -(Dx*Dx+Dy*Dy) / r15
    T2ZZ = (2*Dz*Dz-Dx*Dx-Dy*Dy) / r2
    T3ZZ = (2*Dz*Dz-Dx*Dx-Dy*Dy) / r25
    
    CFEXP = cmath.exp(1j*k*cn*lR)
    
    
    ## setting up the tensor
    xx = CFEXP*(T3XX - ck0*T2XX - k2*T1XX) / eps
    yy = CFEXP*(T3YY - ck0*T2YY - k2*T1YY) / eps
    zz = CFEXP*(T3ZZ - ck0*T2ZZ - k2*T1ZZ) / eps
    
    xy = CFEXP*(T3XY - ck0*T2XY - k2*T1XY) / eps
    xz = CFEXP*(T3XZ - ck0*T2XZ - k2*T1XZ) / eps
    
    yz = CFEXP*(T3YZ - ck0*T2YZ - k2*T1YZ) / eps
        
    yx = xy
    zx = xz
    zy = yz
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy


@numba.njit(cache=True)
def _G0_HE(R1, R2, wavelength, eps):
    """
    R1: dipole position
    R2: evaluation position
    """
    # eps: environment index
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] - R1[2]
    lR2 = (Dx**2 + Dy**2 + Dz**2)
    
    k0 = 2*np.pi / wavelength
    k02n = cmath.sqrt(eps) * k0**2
#-----------------------------------------------------------------
    T2XY = Dz/lR2
    T3XY = Dz/lR2**1.5
#-----------------------------------------------------------------
    T2XZ = -Dy/lR2
    T3XZ = -Dy/lR2**1.5
#-----------------------------------------------------------------
    T2YZ = Dx/lR2
    T3YZ = Dx/lR2**1.5
#-----------------------------------------------------------------
    CFEXP = -1*cmath.exp(1j*k0*cmath.sqrt(eps)*math.sqrt(lR2))
    
    xx = 0
    yy = 0
    zz = 0
    
    xy = CFEXP * (1j*k0*T3XY + k02n*T2XY)
    xz = CFEXP * (1j*k0*T3XZ + k02n*T2XZ)
    yz = CFEXP * (1j*k0*T3YZ + k02n*T2YZ)

    yx = -xy
    zx = -xz
    zy = -yz
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy


## --- "1-2-3" slab propagator, via method of mirror charges for 2 surfaces
@numba.njit(cache=True)
def _G_mirrorcharge_123(R1, R2, wavelength, eps1, eps2, eps3, spacing):
    """
    R1: dipole position
    R2: evaluation position
    """
    Dx = R2[0] - R1[0]
    Dy = R2[1] - R1[1]
    Dz = R2[2] + R1[2]
    
    if R1[2] <= 0:
        cdelta12 = (eps2-eps1)/(eps1+eps2)
        cdelta23 = (eps3-eps2)/(eps3+eps2)
    elif spacing >= R1[2] > 0:
        cdelta12 = (eps1-eps2)/(eps1+eps2)
        cdelta23 = (eps3-eps2)/(eps3+eps2)
    elif R1[2] > spacing:
        cdelta12 = (eps1-eps2)/(eps1+eps2)
        cdelta23 = (eps2-eps3)/(eps3+eps2)
    
    
#!************* Interface: (1,2) *******************************
    r25 = math.pow((Dx*Dx+Dy*Dy+Dz*Dz), 2.5)
     
    SXX12 = cdelta12*(Dz*Dz+Dy*Dy-2*Dx*Dx) / r25
    SYY12 = cdelta12*(Dz*Dz+Dx*Dx-2*Dy*Dy) / r25
    SZZ12 = cdelta12*(2*Dz*Dz-Dx*Dx-Dy*Dy) / r25
    SXY12 = cdelta12*(-3*Dx*Dy) / r25
    SXZ12 = cdelta12*(3*Dx*Dz) / r25
    SYZ12 = cdelta12*(3*Dy*Dz) / r25

#!************* Interface: (2,3) *******************************
    GZ = Dz - 2*spacing
    rgz25 = math.pow((Dx*Dx+Dy*Dy+GZ*GZ), 2.5)
    
    SXX23 = cdelta23*(GZ*GZ+Dy*Dy-2*Dx*Dx) / rgz25
    SYY23 = cdelta23*(GZ*GZ+Dx*Dx-2*Dy*Dy) / rgz25
    SZZ23 = cdelta23*(2*GZ*GZ-Dx*Dx-Dy*Dy) / rgz25
    SXY23 = cdelta23*(-3*Dx*Dy) / rgz25
    SXZ23 = cdelta23*(3*Dx*GZ) / rgz25
    SYZ23 = cdelta23*(3*Dy*GZ) / rgz25
#!**************************************************************
    
    xx = (SXX12+SXX23)
    yy = (SYY12+SYY23)
    zz = (SZZ12+SZZ23)
    
    xy = (SXY12+SXY23)
    xz = (SXZ12+SXZ23)
    
    yx = xy
    yz = (SYZ12+SYZ23)
    
    zx = -xz
    zy = -yz
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy



## --- the full propagator: vacuum + surface term -- nearfield approximation!
##     assuming emitter in center layer, to calcualte resulting field anywhere
@numba.njit(cache=True)
def G0_EE_123(R1, R2, wavelength, eps1, eps2, eps3, spacing):
    """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
    R1: dipole position
    R2: evaluation position
    """
    if eps1==eps2 and eps2==eps3:
        eps = eps2
        scr_f = 1.0
    else:
        ## ----- emitter dipole situated in substrate layer ('eps1')
        if R1[2] < 0:
            ## field evaluated in bottom layer ("substrate")
            if R2[2] < 0:
                eps = eps1
                scr_f = 1.0
            ## field evaluated in center layer (same as emitter dipole)
            elif 0 <= R2[2] <= spacing:
                eps = eps2
                scr_f = eps * 2.0/(eps1+eps2)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in top cladding layer
            elif spacing < R2[2]:
                eps = eps3
                scr_f = eps * 2.0/(eps1+eps3)   # factor 'eps' to take 1/eps in G0 into account
        
        ## ----- emitter dipole situated in center layer ('eps2')
        if 0 <= R1[2] <= spacing:
            ## field evaluated in bottom layer ("substrate")
            if R2[2] < 0:
                eps = eps1
                scr_f = eps * 2.0/(eps1+eps2)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in center layer (same as emitter dipole)
            elif 0 <= R2[2] <= spacing:
                eps = eps2
                scr_f = 1.0
            ## field evaluated in top cladding layer
            elif spacing < R2[2]:
                eps = eps3
                scr_f = eps * 2.0/(eps2+eps3)   # factor 'eps' to take 1/eps in G0 into account
        
        ## ----- emitter dipole situated in top cladding layer ('eps3')
        if R1[2] > spacing:
            ## field evaluated in bottom layer ("substrate")
            if R2[2] < 0:
                eps = eps1
                scr_f = eps * 2.0/(eps1+eps3)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in center layer (same as emitter dipole)
            elif 0 <= R2[2] <= spacing:
                eps = eps2
                scr_f = eps * 2.0/(eps2+eps3)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in top cladding layer
            elif spacing < R2[2]:
                eps = eps3
                scr_f = 1.0
    
    ## ----- free space dyad
    xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0(R1, R2, wavelength, eps)

    return scr_f*xx, scr_f*yy, scr_f*zz, \
           scr_f*xy, scr_f*xz, scr_f*yx, \
           scr_f*yz, scr_f*zx, scr_f*zy


@numba.njit(cache=True)
def Gs_EE_123(R1, R2, wavelength, eps1, eps2, eps3, spacing):
    """evaluate field of mirror-charge in 1-2-3 layer environment (NF approx.)
    
    R1: dipole position
    R2: evaluation position
    """
    xx, yy, zz, xy, xz, yx, yz, zx, zy = [np.complex64(0)]*9
    
    ## ----- actual interface present
    if eps1!=eps2 or eps2!=eps3:
        ## ----- emitter dipole situated in same layer as observer
        if ( (0 <= R1[2] <= spacing and 0 <= R2[2] <= spacing) or
             (R1[2] < 0 and R2[2] < 0) or 
             (R1[2] > spacing and R2[2] > spacing) ):
            ## mirror charges NF Green's tensor
            xx, yy, zz, xy, xz, yx, yz, zx, zy = _G_mirrorcharge_123(R1, R2, wavelength, 
                                                        eps1, eps2, eps3, spacing)

    return xx, yy, zz, xy, xz, yx, yz, zx, zy


@numba.njit(cache=True)
def Gtot_EE_123(R1, R2, wavelength, eps1, eps2, eps3, spacing):
# def G_EE_123(R1, R2, wavelength, conf_dict):
    """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
    R1: dipole position
    R2: evaluation position
    """
    ## ----- free space term
    xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_EE_123(R1, R2, wavelength, 
                                                   eps1, eps2, eps3, spacing)
    
    ## ----- surface term
    xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_EE_123(R1, R2, wavelength, 
                                                    eps1, eps2, eps3, spacing)

    return xx + xxs, yy + yys, zz + zzs, \
           xy + xys, xz + xzs, yx + yxs, \
           yz + yzs, zx + zxs, zy + zys



@numba.njit(cache=True)
def G_HE_123(R1, R2, wavelength, eps1, eps2, eps3, spacing=0):
    """evaluate field of dipole emitter in 1-2-3 layer environment
    
    R1: dipole position
    R2: evaluation position
    """
    ## field evaluated in bottom layer ("substrate")
    if R2[2] < 0:
        eps = eps1
    ## field evaluated in center layer (same as emitter dipole)
    elif 0 <= R2[2] <= spacing:
        eps = eps2
    ## field evaluated in top cladding layer
    elif spacing <= R2[2]:
        eps = eps3
    
    ## ----- free space contribution
    xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0_HE(R1, R2, wavelength, eps)
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy





### --- ff asymptotic 3D
# =============================================================================
# asymptotic propagators for far-field, 3D
# =============================================================================
@numba.njit(cache=True)
def _G0_EE_asymptotic(R1, R2, wavelength, eps2): 
    """Asymptotic vacuum Green's Dyad
    
    Parameters
    ----------
    R1 : np.array of 3 float
        dipole position [x,y,z]
    R2 : np.array of 3 float
        evaluation position
    wavelength : float
        emitter wavelength in nm
    eps1 : float, complex
        substrate permittivity
    """
    ## transform to spherical coordinates
    lR = np.linalg.norm(R2)
    theta = np.arccos(R2[2] / lR)
    phi = np.arctan2(R2[1], R2[0])
    
    if lR < wavelength:
        raise Exception("Distance too close. Trying to evaluate asymtpotic far-field dyad in the near-field region (R < wavelength).")
    
    ## wavenumber
    kvac = 2 * np.pi / wavelength                   # in vacuum
    k2 = 2 * np.pi * cmath.sqrt(eps2) / wavelength     # in surrounding medium of emitting dipole
    
    ## tensor prefactor
    A = ((kvac**2)*np.exp(1.0j*k2*lR)/lR * 
                      np.exp(-1.0j*k2*np.sin(theta) * 
                     (R1[0]*np.cos(phi) + R1[1]*np.sin(phi))) * 
                      np.exp(-1.0j*k2*np.cos(theta)*R1[2]))
    
    ## matrix elements
    S0_XX = A * (1.-np.sin(theta)**2*np.cos(phi)**2)
    S0_XY = A * (-np.sin(theta)**2*np.cos(phi)*np.sin(phi))
    S0_XZ = A * (-np.sin(theta)*np.cos(theta)*np.cos(phi))
    S0_YX = A * (-np.sin(theta)**2*np.cos(phi)*np.sin(phi))
    S0_YY = A * (1.-np.sin(theta)**2*np.sin(phi)**2)
    S0_YZ = A * (-np.sin(theta)*np.cos(theta)*np.sin(phi))
    S0_ZX = A * (-np.sin(theta)*np.cos(theta)*np.cos(phi))
    S0_ZY = A * (-np.sin(theta)*np.cos(theta)*np.sin(phi))
    S0_ZZ = A * (np.sin(theta)**2)
    
    return S0_XX, S0_YY, S0_ZZ, \
           S0_XY, S0_XZ, S0_YX, \
           S0_YZ, S0_ZX, S0_ZY


@numba.njit(cache=True, fastmath=True)
def Gs_EE_asymptotic(R1, R2, wavelength, eps1, eps2, eps3=0, spacing=0):
    """Asymptotic Green's Dyad for dipole above dielectric interface (inside eps2)
    
    Electric field propagator for electric dipole transition.
    
    In contrast to the nearfield propagators, this includes the vacuum contribution!
    
    Parameters
    ----------
    R1 : np.array of 3 float
        dipole position [x,y,z]
    R2 : np.array of 3 float
        evaluation position
    wavelength : float
        emitter wavelength in nm
    eps1 : float, complex
        substrate permittivity
    eps2 : float, complex
        emitter environment permittivity (above substrate)
    eps3, spacing : dummy parameters, ignored
    
    """
    ## transform to spherical coordinates
    lR = np.linalg.norm(R2)
    theta = np.arccos(R2[2] / lR)
    phi = np.arctan2(R2[1], R2[0])

    
    if lR < wavelength:
        raise Exception("Distance too close. Trying to evaluate asymtpotic far-field dyad in the near-field region (R < wavelength).")
    
    ## refractive index of media
    n1 = cmath.sqrt(eps1)
    n2 = cmath.sqrt(eps2)
    ## wavenumber
    kvac = 2 * np.pi / wavelength        # in vacuum
    k1 = 2 * np.pi * n1 / wavelength     # in substrate
    k2 = 2 * np.pi * n2 / wavelength     # in surrounding medium of emitting dipole
    
    ## workaround for positions too close to substrate
    if np.abs(np.cos(theta)) < 1E-10:
        theta -= np.sign(np.cos(theta))*0.001*np.pi
    
# =============================================================================
# Surface propagator above the surface
# =============================================================================
    if np.cos(theta) >= 0:
        ## tensor prefactors
        A0 = (kvac**2 * np.exp(1.0j*k2*lR) / lR * 
                              np.exp(-1.0j*k2*np.sin(theta) * 
                             (R1[0]*np.cos(phi) + R1[1]*np.sin(phi))))
        A_vac = A0 * np.exp(-1.0j*k2*np.cos(theta)*R1[2])
        A_up = -1 * A0 * np.exp(1.0j*k2*np.cos(theta)*R1[2])
        
        
        ## ------------------------------ vacuum contribution
        ## matrix elements
        S0_11 = A_vac * (1.-np.sin(theta)**2 * np.cos(phi)**2)
        S0_12 = A_vac * (-np.sin(theta)**2 * np.cos(phi) * np.sin(phi))
        S0_13 = A_vac * (-np.sin(theta) * np.cos(theta) * np.cos(phi))
        S0_21 = A_vac * (-np.sin(theta)**2 * np.cos(phi) * np.sin(phi))
        S0_22 = A_vac * (1.-np.sin(theta)**2 * np.sin(phi)**2)
        S0_23 = A_vac * (-np.sin(theta) * np.cos(theta) * np.sin(phi))
        S0_31 = A_vac * (-np.sin(theta) * np.cos(theta) * np.cos(phi))
        S0_32 = A_vac * (-np.sin(theta) * np.cos(theta) * np.sin(phi))
        S0_33 = A_vac * (np.sin(theta)**2)
        
        
        ## ------------------------------ surface contribution
        ## -- Fresnel coefficients
        r_p = ((eps1*n2*np.cos(theta) - eps2*cmath.sqrt(eps1 - eps2*np.sin(theta)**2)) /
               (eps1*n2*np.cos(theta) + eps2*cmath.sqrt(eps1 - eps2*np.sin(theta)**2)))
        
        r_s = ((n2*np.cos(theta) - cmath.sqrt(eps1 - eps2*np.sin(theta)**2)) /
               (n2*np.cos(theta) + cmath.sqrt(eps1 - eps2*np.sin(theta)**2)))
        
        ## -- matrix elements
        Sp11 = A_up * (r_p * np.cos(theta)**2 * np.cos(phi)**2 - r_s*np.sin(phi)**2)
        Sp12 = A_up * ((r_p * np.cos(theta)**2 + r_s) * np.sin(phi) * np.cos(phi))
        Sp13 = A_up * (r_p * np.cos(theta) * np.sin(theta) * np.cos(phi))
        Sp21 = Sp12
        Sp22 = A_up * (r_p * np.cos(theta)**2*np.sin(phi)**2 - r_s*np.cos(phi)**2)
        Sp23 = A_up * (r_p * np.cos(theta) * np.sin(theta) * np.sin(phi))
        Sp31 = -Sp13
        Sp32 = -Sp23
        Sp33 = A_up * (-r_p * np.sin(theta)**2)
        
        Sp11 += S0_11
        Sp12 += S0_12
        Sp13 += S0_13
        Sp21 += S0_21
        Sp22 += S0_22
        Sp23 += S0_23
        Sp31 += S0_31
        Sp32 += S0_32
        Sp33 += S0_33
    
        
# =============================================================================
# Surface propagator under the surface 
# =============================================================================
    else:
        ## -- coeff.
        D_eps_eff = cmath.sqrt(eps2 - eps1*np.sin(theta)**2)
        
        delta_s = ((-n1*np.cos(theta) - D_eps_eff)/
                   (-n1*np.cos(theta) + D_eps_eff))
        tau_s = 1. - delta_s
        phi_s = n1*tau_s / D_eps_eff 
        
        delta_p = ((-eps2*n1*np.cos(theta) - eps1*D_eps_eff)/
                   (-eps2*n1*np.cos(theta) + eps1*D_eps_eff))
        tau_p = delta_p + 1.
        phi_p = n1*tau_p * D_eps_eff
    
        A_low = ((kvac**2)*np.exp(1.0j*k1*lR) / lR*np.exp(-1.0j*k1*np.sin(theta)*
              (R1[0]*np.cos(phi) + R1[1]*np.sin(phi)))*np.exp(1.0j*kvac*D_eps_eff*R1[2]))
        
        ## -- matrix elements
        Sp11 = A_low * ((phi_p/eps2*np.cos(phi)**2 + phi_s*np.sin(phi)**2)*np.cos(theta))
        Sp12 = A_low * ((phi_p/eps2-phi_s)*np.cos(theta)*np.sin(phi)*np.cos(phi))
        Sp13 = A_low * (tau_p*eps1/eps2*np.cos(phi)*np.cos(theta)*np.sin(theta))
        Sp21 = A_low * ((phi_p/eps2-phi_s)*np.cos(theta)*np.sin(phi)*np.cos(phi))
        Sp22 = A_low * ((phi_p/eps2*np.sin(phi)**2 + phi_s*np.cos(phi)**2)*np.cos(theta))
        Sp23 = A_low * (tau_p*eps1/eps2*np.sin(phi)*np.cos(theta)*np.sin(theta))
        Sp31 = A_low * (-phi_p/eps2*np.sin(theta)*np.cos(phi))
        Sp32 = A_low * (-phi_p/eps2*np.sin(theta)*np.sin(phi))
        Sp33 = A_low * (-tau_p*eps1/eps2*np.sin(theta)**2)
    
    return Sp11, Sp22, Sp33, \
           Sp12, Sp13, Sp21, \
           Sp23, Sp31, Sp32





# =============================================================================
# ensemble-propagator routines
# =============================================================================
## --- multi-dipole / multi-probe propagator evaluation
@numba.njit(parallel=True)
def greens_tensor_evaluation(dp_pos, r_probe, G_func, wavelength, conf_dict, M, 
                             selfterm=np.zeros((3,3)).astype(np.complex64), 
                             dist_div_G=0.001):
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    for i in numba.prange(len(dp_pos)):   # explicit parallel loop
        _pos = dp_pos[i]
        for j in range(len(r_probe)):
            _r = r_probe[j]
            if np.sqrt((_r[0]-_pos[0])**2 + (_r[1]-_pos[1])**2 + (_r[2]-_pos[2])**2)<dist_div_G:
                xx, xy, xz = selfterm[0]
                yx, yy, yz = selfterm[1]
                zx, zy, zz = selfterm[2]
                # xx, yy, zz, xy, xz, yx, yz, zx, zy = selfterm
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = G_func(_pos, _r, wavelength, 
                                                            eps1, eps2, eps3, spacing)
            ## return list of Greens tensors
            M[i,j,0,0], M[i,j,1,1], M[i,j,2,2] = xx, yy, zz
            M[i,j,1,0], M[i,j,2,0], M[i,j,0,1] = yx, zx, xy
            M[i,j,2,1], M[i,j,0,2], M[i,j,1,2] = zy, xz, yz

                
@numba.njit(parallel=True, cache=True)
def t_sbs_EE_123_quasistatic(geo, wavelength, selfterms, alpha, conf_dict, M):
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]       # "observer"
        for j in range(len(geo)):
            R1 = geo[j]   # emitter
            aj = alpha[j]
            st = selfterms[j]
            ## --- vacuum dyad
            if i==j:
                ## self term
                xx, yy, zz = st[0,0], st[1,1], st[2,2]
                xy, xz, yx = st[0,1], st[0,2], st[1,0]
                yz, zx, zy = st[1,2], st[2,0], st[2,1]
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_EE_123(R1, R2, wavelength, 
                                                                eps1, eps2, eps3, spacing)
            
            ## --- 1-2-3 surface dyad (non retarded NF approximation)
            if eps1!=eps2 or eps2!=eps3:
                xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_EE_123(
                                  R1, R2, wavelength, eps1, eps2, eps3, spacing)
                ## combined dyad
                xx, yy, zz, xy, xz, yx, yz, zx, zy = xx+xxs, yy+yys, zz+zzs, \
                                                      xy+xys, xz+xzs, yx+yxs, \
                                                      yz+yzs, zx+zxs, zy+zys
            
            ## return invertible matrix:  delta_ij*1 - G[i,j] * alpha[j]
            M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
            M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
            M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
            M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
            M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
            M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
            M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
            M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
            M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])
            if i==j:
                M[3*i+0, 3*j+0] += 1
                M[3*i+1, 3*j+1] += 1
                M[3*i+2, 3*j+2] += 1


@numba.njit(parallel=True, cache=True)
# def t_sbs_HE_123_quasistatic(geo, wavelength, eps1, eps2, eps3, spacing, selfterms, alpha, M):
def t_sbs_HE_123_quasistatic(geo, wavelength, selfterms, alpha, conf_dict, M):
    eps2 = conf_dict['eps2']
    
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]        # "observer"
        for j in range(len(geo)):
            R1 = geo[j]    # emitter
            aj = alpha[j]
            st = selfterms[j]
            ## --- vacuum dyad
            if i==j:
                ## self term
                xx, yy, zz = st[0,0], st[1,1], st[2,2]
                xy, xz, yx = st[0,1], st[0,2], st[1,0]
                yz, zx, zy = st[1,2], st[2,0], st[2,1]
            else:
                ## we need G^HE: H-field due to e-dipole
                xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0_HE(R1, R2, wavelength, eps2)
            
            ## return: G[i,j] * alpha[j]
            ## --- magnetic-electric part
            M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
            M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
            M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
            M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
            M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
            M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
            M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
            M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
            M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])




### --- dyad class 3D non-retarded
class DyadsQuasistatic123(DyadsBaseClass):
    __name__ = "3-layer environment: quasistatic 3D '1-2-3' Green's tensors"
    
    def __init__(self, n1=None, n2=None, n3=None, spacing=5000, 
                 radiative_correction=True):
        super().__init__()
        
        ## Dyads
        self.G_EE = Gtot_EE_123
        self.G_HE = G_HE_123
        self.G_EE_ff = Gs_EE_asymptotic
        
        ## evaluate propagator routine
        self.eval_G = greens_tensor_evaluation
        
        ## coupling matrix constructor routines
        self.tsbs_EE = t_sbs_EE_123_quasistatic
        self.tsbs_HE = t_sbs_HE_123_quasistatic
        
        ## environment definition
        ## set ref. index values or material class of environment layers
        from pyGDM2 import materials
        if isinstance(n1, (int, float, complex)) and not isinstance(n1, bool):
            self.n1_material = materials.dummy(n1)
        else:
            self.n1_material = n1
            
        n2 = n2 or n1     # if None, use `n1`
        if isinstance(n2, (int, float, complex)) and not isinstance(n2, bool):
            from pyGDM2 import materials
            self.n2_material = materials.dummy(n2)
        else:
            self.n2_material = n2
            
        n3 = n3 or n2     # if None, use `n2`
        if isinstance(n3, (int, float, complex)) and not isinstance(n3, bool):
            from pyGDM2 import materials
            self.n3_material = materials.dummy(n3)
        else:
            self.n3_material = n3
        
        self.spacing = spacing
        self.radiative_correction = radiative_correction
    
    def __repr__(self, verbose=False):
        """description about simulation environment defined by set of dyads
        """
        out_str =  ' ------ environment -------'
        out_str += '\n ' + self.__name__
        out_str += '\n '
        out_str += '\n' + '   n3 = {}  <-- top'.format(self.n3_material.__name__)
        out_str += '\n' + '   n2 = {}  <-- structure zone (height "spacing" = {}nm)'.format(
                        self.n2_material.__name__, self.spacing)
        out_str += '\n' + '   n1 = {}  <-- substrate'.format(self.n1_material.__name__)
        return out_str
    
    
    def exceptionHandling(self, struct, efield):
        """Exception handling / consistency check for the set of tensors
        
        check if structure and incident field generator are compatible

        Parameters
        ----------
        struct : :class:`.structures.struct`
            instance of structure class
        efield : :class:`.fields.efield`
            instance of incident field class

        Returns
        -------
        bool : True if struct and field are compatible, False if they don't fit the tensors

        """
        if len(struct.geometry) > 0:
            z_min = struct.geometry.T[2].min()
            z_max = struct.geometry.T[2].max()
            ## if interface 1/2 exists, check if entire structure is below or above
            if self.n1_material.__name__ != self.n2_material.__name__:
                if z_min<0 and z_max>0:
                    warnings.warn("Structure in-between substrate and middle-layer. " +
                                  "This is not supported and will most likely falsify the simulation.")
            
            ## if interface 2/3 exists, check if entire structure is below or above
            if self.n2_material.__name__ != self.n3_material.__name__:
                if z_min<self.spacing and z_max>self.spacing:
                    warnings.warn("Structure in-between middle and top cladding layer. " +
                                  "This is not supported and will most likely falsify the simulation.")
            
        return True
    
    
    def getConfigDictG(self, wavelength, struct, efield):
        ## all data need to be same dtype, must be cast to correct type inside numba functions
        conf_dict = numba.typed.Dict.empty(key_type=numba.types.unicode_type,
                                           value_type=numba.types.complex64)
        
        conf_dict['eps1'] = np.complex64(self.n1_material.epsilon(wavelength))
        conf_dict['eps2'] = np.complex64(self.n2_material.epsilon(wavelength))
        conf_dict['eps3'] = np.complex64(self.n3_material.epsilon(wavelength))
        conf_dict['spacing'] = np.complex64(self.spacing)
        
        ## return a numba typed dictionary of "complex64" type,
        ## can be used to pass configuration to the green's functions
        return conf_dict
    
    
    def getEnvironmentIndices(self, wavelength, geo):
        """get environment permittivity for `wavelength` at each meshpoint"""
        self.n1 = self.n1_material.epsilon(wavelength)**0.5
        self.n2 = self.n2_material.epsilon(wavelength)**0.5
        self.n3 = self.n3_material.epsilon(wavelength)**0.5
        
        ## environment epsilon at every meshpoint
        eps_env = np.zeros(len(geo), dtype=self.dtypec)
        eps_env[geo.T[2] > self.spacing] = self.n3_material.epsilon(wavelength)
        eps_env[np.logical_and(0 <= geo.T[2], geo.T[2] <= self.spacing)] = self.n2_material.epsilon(wavelength)
        eps_env[geo.T[2] < 0] = self.n1_material.epsilon(wavelength)
        
        return eps_env
        
        
    def getSelfTermEE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        k0 = 2.0*np.pi / float(wavelength)
        
        if struct.normalization == 0:
            cnorm = np.zeros(len(eps_env))
        else:
            norm_nonrad = -4.0 * np.pi * struct.normalization / (3.0 * struct.step**3 * eps_env)
            
            if self.radiative_correction:
                norm_rad = 1j * 2.0 * struct.normalization * (k0**3)/3.0 * np.ones(len(norm_nonrad))
                cnorm = norm_nonrad + norm_rad
            else:
                cnorm = norm_nonrad
        
        self_term_tensors_EE = np.zeros([len(norm_nonrad), 3, 3], dtype=self.dtypec)
        self_term_tensors_EE[:,0,0] = cnorm
        self_term_tensors_EE[:,1,1] = cnorm
        self_term_tensors_EE[:,2,2] = cnorm
        
        return self_term_tensors_EE
        
    
    def getSelfTermHE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        self_term_tensors_HE = np.zeros([len(eps_env), 3, 3], dtype=self.dtypec)
        
        return self_term_tensors_HE
        
        
    def getPolarizabilityTensor(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        normalization = struct.normalization
        eps = struct.epsilon_tensor 
        
        vcell_norm = struct.step**3 / float(normalization)
        
        eps_env_tensor = np.zeros(eps.shape, dtype=self.dtypec)
        eps_env_tensor[:,0,0] = eps_env
        eps_env_tensor[:,1,1] = eps_env
        eps_env_tensor[:,2,2] = eps_env
        
        ## --- isotropic polarizability
        alphatensor = (eps - eps_env_tensor) * vcell_norm / (4.0 * np.pi)

        return alphatensor





